#include "AFMM.h"
#include <math.h>
#include <iostream>
#include <stack>


using namespace std;


struct  NewValue {  int i; int j; float value; }; 




AugmentedFastMarchingMethod::AugmentedFastMarchingMethod(Image<float>* f_,FLAGS* flags_,Image<float>* count_,
						       Image<std::multimap<float,int> >* origins_)
			   :FastMarchingMethod(f_,flags_),count(count_),scan_dir(0),comp_method(AFMM_STAR),
			    origins(origins_), dist_tol(sqrt(2.0))
{
}




void AugmentedFastMarchingMethod::init_count()
{
   count->makeImage(f->getWidth(),f->getHeight());			//0. Resize count[] to match the size of our images
   
   count->clearImage(FLAGS::MYINFINITY);			
   icount->clearImage(FLAGS::MYINFINITY);					//1. Collect all narrowband points in a container (s)
   stack<Coord> s; int i,j;									//   We will next label these points so that neighbor points
															//   have consecutive labels (in icount[]). 
   if(scan_dir==0)											//   Additionally, in count[], we will construct a boundary
   { for(j=0;j<count->getHeight();j++)						//   arc-length parameterization. The value infinity marks		
       for(i=0;i<count->getWidth();i++)						//   points which are not (yet) labeled.
          if (flags->narrowband(i,j)) s.push(Coord(i,j));	//   
   }
   else
     for(j=count->getHeight()-1;j>=0;j--)			
       for(i=0;i<count->getWidth();i++) 
          if (flags->narrowband(i,j)) s.push(Coord(i,j));
	
   length = 0; 						
   int  C = 0;	
	
   const int MYINFINITY_2 = FLAGS::MYINFINITY/2;			//2. Now that we collected all narrowband points,
   for(float k=0;s.size();)									//   compute the count and icount fields.	
   {
      Coord pt = s.top(); s.pop();							//2.1. Pick narrowband-point, test if already numbered
      if ((*count)(pt.j,pt.i)<MYINFINITY_2) continue;		//     If so, nothing to do for it.

      int ci=0,cj=0;										//2.2. Point 'pt' not numbered, we'll number it next
	  float min_c = FLAGS::MYINFINITY;
	  
      for(int dj=-1;dj<2;dj++)								//2.3. Find neighbor of 'pt' with minimal label value
		for(int di=-1;di<2;di++)
		{  
			int ii=pt.i+di; int jj=pt.j+dj; 
			if (ii<0 || ii>=count->getWidth() || jj<0 || jj>=count->getHeight() || !flags->narrowband(ii,jj)) continue;         
			if ((*count)(jj,ii)>=min_c) continue;
			ci = ii; cj = jj; min_c = (*count)(jj,ii);
		}
	   
	  float d  = sqrt(float((pt.i-ci)*(pt.i-ci)+(pt.j-cj)*(pt.j-cj)));
	  float c  = (min_c>MYINFINITY_2)? k+1 : min_c+d;		//2.4. Compute label 'c' of 'pt'. If there's a labeled neighbor,
															//     then we simply add the distance from 'pt' to it to that label.
															//     If not, it means we've found the 1st point of a new boundary segment,
															//	   so we use, as label, the length of the boundary-so-far plus 1.

      (*count)(pt.j,pt.i)  = c;								//2.5. Update both the boundary-labels (icount) and boundary-parameterization (count)
      (*icount)(pt.j,pt.i) = C;
      from[C] = pt;											//2.6. Register all boundary-points in from[]
      if (origins)											//If we compute the full-FT, initialize full-FT for boundary-pts
		  (*origins)(pt.j,pt.i).insert(std::make_pair(0,C));//to the identity function (a boundary-point is its own feature-point)
      C++;													//2.7. One more boundary point

      if (c>length) length=c;								//2.8. Update maximal boundary-parameter value
      if (k<c) k=c;											//

      for(int dj=-1;dj<2;dj++)								//2.9. Search for new boundary points around the just-labeled point
		for(int di=-1;di<2;di++)							//     and add them to our stack. This ensures that we next label points
		{													//     in the order they come on the boundary, consecutively.
			if (di==0 && dj==0) continue;
			int ii=pt.i+di; int jj=pt.j+dj; 
			if (ii<0 || ii>=count->getWidth() || jj<0 || jj>=count->getHeight() || (*count)(jj,ii)<MYINFINITY_2 || !flags->narrowband(ii,jj)) 
				continue;  		
			s.push(Coord(ii,jj));
        }	
   }
}




int AugmentedFastMarchingMethod::execute(float max_distance,int max_iterations)
{															//Compute the one-point or full-point feature transforms

   from.resize(flags->initialContourLength());				//Allocate memory for from[]
   
   icount = new Image<int>(f->getWidth(),f->getHeight());	//Allocate memory for icount[]

   if (comp_method!=AFMM_VDT) origins = 0;					//If we don't compute the full-FT, we don't use origins[] at all
   

   init_count();											//Initialize the count and icount fields on boundary

   int ret = FastMarchingMethod::execute(max_distance,max_iterations);
															//Call inherited fast-marching-method. This traverses all points
															//in increasing distance-to-boundary. Our overloaded add_to_narrowband()
															//and diffuse() methods next compute our feature transforms

   delete icount;											//Ready with icount

   for(int j=0;j<count->getHeight();++j)					//Set count for all points which were not traversed (i.e., outside of
   for(int i=0;i<count->getWidth();++i)						//the input shape) to -1. This allows one to visualize the field 
   {														//count without messing with infinity values
      float& v = (*count)(j,i);
	  if (v==FLAGS::MYINFINITY) v = -1;
   }	  

   return ret;												//Return the number of iterations of the FMM
}




void AugmentedFastMarchingMethod::add_to_narrowband(int i,int j,int active_i,int active_j)
{															//Overloads the method inherited from FastMarchingMethod to
															//compute the full-feature-transform

	FastMarchingMethod::add_to_narrowband(i,j,active_i,active_j);
															//Call inherited (base) method
	if (comp_method!=AFMM_VDT) return;						//If we don't compute full-FT's, nothing more to do

	std::multimap<float,int> om;							//Candidates for origins of (i,j)

	float MYINFINITY_2 = FLAGS::MYINFINITY/2; 

	for(int ii=i-1;ii<=i+1;ii++)							//For all known neighbors of current point (i,j):		
	for(int jj=j-1;jj<=j+1;jj++)
	if (flags->alive(ii,jj) && (*count)(jj,ii)<MYINFINITY_2)
	{						
		 std::multimap<float,int>& o = (*origins)(jj,ii);	//All origins of neighbor (ii,jj)
		 for(std::multimap<float,int>::iterator oi=o.begin();oi!=o.end();oi++)
		 {
		   int   O = (*oi).second;
		   Coord c = from[O];								//Find where neighbor came from on the initial boundary
		   int   d = (c.i-i)*(c.i-i)+(c.j-j)*(c.j-j);		//Find distance from current point (i,j) to its potential origin
		 
		   bool go=true;									//Collect (i,j)'s potential origins and their distances
		   for(std::multimap<float,int>::iterator it=om.begin();go && it!=om.end();it++)
			 if ((*it).second==O) go=false;					//Take care we don't insert same origin twice
		   if (go)
			 om.insert(make_pair(sqrt(float(d)),O));
	 }
	}

															//Update origins of (i,j)
															//These are simply all points in our potential-list om[]
	float dmin = (*om.begin()).first + dist_tol;			//with smallest distance from the crt-point, using dist_tol.
	int pt_i = from[(*om.begin()).second].i;				//Since om[] is sorted on distance, this means its first elements
	int pt_j = from[(*om.begin()).second].j;				//until we reach (*om.begin()).first + dist_tol
	std::multimap<float,int>& oa = (*origins)(j,i); 
	for(std::multimap<float,int>::iterator oi=om.begin();oi!=om.end() && (*oi).first<dmin;oi++)
		oa.insert(*oi); 
}




int AugmentedFastMarchingMethod::diffuse()					//Replace inherited method to compute both the distance-transform
{															//and the one-point feature transform
    static NewValue newp[20]; 

    int min_i,min_j;										//*** 1. FIND MIN-POINT IN NARROWBAND WITH LOWEST VALUE	
    std::multimap<float,Coord>::iterator it=map.begin();
    if (it==map.end()) return 0;							//

    min_i  = it->second.i;
    min_j  = it->second.j;
    map.erase(it);											//erase point from 'map', since we'll make it alive in step 2

															//*** 2. MAKE MIN-POINT ALIVE; Its distance value is now final
    (*flags)(min_j,min_i) = FLAGS::ALIVE;		
    if ((*f)(min_j,min_i)>=maxf) return 1;					//stop evolution if we reached the user-prescribed distance threshold

    Coord nbs[4]; int nn = 0;								//*** 3. FIND ALL ITS STILL-TO-BE-UPDATED NEIGHBOURS
    tag_nbs(min_i-1,min_j,min_i,min_j,nbs,nn);
    tag_nbs(min_i+1,min_j,min_i,min_j,nbs,nn);
    tag_nbs(min_i,min_j-1,min_i,min_j,nbs,nn);
    tag_nbs(min_i,min_j+1,min_i,min_j,nbs,nn);

    if (!nn)												//no more alive-neighbors of point (min_i,min_j) found,
    {														//so it should be an extremum point...
        nextr++;
    }

															//*** 4. UPDATE VALUES OF NEIGHBOURS OF MIN-POINT
    NewValue* nnewp = newp;									//start updating neighbours. Their new values will be saved
    for(nn--;nn>=0;nn--)									//in nnewp[] and pasted back in 'f' at the update end.
    {
		int i = nbs[nn].i, j = nbs[nn].j;

		float vi_1j = (*f)(j,i-1),     vijx1 = (*f)(j+1,i);
		float vix1j = (*f)(j,i+1),     vij_1 = (*f)(j-1,i);
		int   fi_1j = (*flags)(j,i-1), fijx1 = (*flags)(j+1,i);
		int   fix1j = (*flags)(j,i+1), fij_1 = (*flags)(j-1,i);

		float sol = FLAGS::MYINFINITY; 
		solve(fi_1j,fij_1,vi_1j,vij_1,sol);
		solve(fix1j,fij_1,vix1j,vij_1,sol);
		solve(fi_1j,fijx1,vi_1j,vijx1,sol); 
		solve(fix1j,fijx1,vix1j,vijx1,sol); 

		if (sol < FLAGS::MYINFINITY/2) 
		{ nnewp->i = i; nnewp->j = j; nnewp->value = sol; ++nnewp; }
    }

	--nnewp;												//***5. Write updated values back in field.
    for(NewValue* x=nnewp;x>=newp;--x)						//for all updated neighbours:
    {
       map.erase(ptrs(x->j,x->i));							//remove the neighbour's entry from the sorted map...
															//...and insert it back since its field-value changed
       ptrs(x->j,x->i) = map.insert(make_pair(x->value,Coord(x->i,x->j)));						
       (*f)(x->j,x->i) = x->value;							//update the field too
    }
		
	if (comp_method==AFMM_STAR)								//AFMM_STAR method: update count[] and icount[] for all points with new distances
	for(NewValue* x=nnewp;x>=newp;--x)
	{
		int dmin = FLAGS::MYINFINITY; int pt_i,pt_j;	
		for(int ii=x->i-1;ii<=x->i+1;++ii)					//For all known neighbors of crt point (i,j):		
			for(int jj=x->j-1;jj<=x->j+1;++jj)
				if (!flags->faraway(ii,jj) && (*count)(jj,ii)<FLAGS::MYINFINITY/2)
				{
					int   V = (*icount)(jj,ii);				//Find where (ii,jj) came from on the initial boundary
					const Coord& c = from[V];
					int   d = (c.i-x->i)*(c.i-x->i)+(c.j-x->j)*(c.j-x->j);
					if (d<dmin)					
					{  dmin=d; pt_i=ii; pt_j=jj; }			//Find boundary-point closest to (ii,jj)
				}
	   	
		(*count)(x->j,x->i)  = (*count)(pt_j,pt_i);
        (*icount)(x->j,x->i) = (*icount)(pt_j,pt_i);									
	}
	
    return 1;
}


