#include "FMM.h"
#include <math.h>
#include <iostream>


using namespace std;


struct 	NewValue {  int i; int j; float value;  };	//Used in the diffuse() routine



FastMarchingMethod::FastMarchingMethod(Image<float>* f_,FLAGS* flags_)
		   :f(f_),flags(flags_),ptrs(f_->getWidth(),f_->getHeight()),verbose(false)
{
   for(int j=0;j<flags->getHeight();j++)
      for(int i=0;i<flags->getWidth();i++)
	 if (flags->narrowband(i,j))
	 {
	    std::multimap<float,Coord>::value_type v((*f)(j,i),Coord(i,j));
	    ptrs(j,i) = map.insert(v);
	 } 
}


FastMarchingMethod::~FastMarchingMethod()
{  }


int FastMarchingMethod::execute(float max_distance,int max_iterations)
{
   int cc,i;

   maxf = max_distance;
   N    = max_iterations;

   int step = flags->initialObjectSize()/10;   

   for(i=0,cc=0;i<N;i++,cc++)					//Iterate until max # iterations done or all image is processed
   {
      if (!diffuse()) break;
      if (verbose && cc==step)
      { cout<<"Iteration "<<i<<" done"<<endl; cc=0; }
   }

   return i;
}




void FastMarchingMethod::solve(int fi_1j,int fij_1,float vi_1j,float vij_1,float& sol)
{
	if (fi_1j == FLAGS::ALIVE || fi_1j == FLAGS::EXTREMUM)
	  if (fij_1 == FLAGS::ALIVE || fij_1 == FLAGS::EXTREMUM)	//sol determined by two points: solve order-2 equation
	  {
	     float d = 2 - (vi_1j-vij_1)*(vi_1j-vij_1);
	     if (d>=0)
	     {	
	       float r   = sqrt(d);
	       float ss  = ((vi_1j+vij_1) - r)/2; 
		   if (ss >= vi_1j && ss >= vij_1 && ss<sol) sol = ss;
	       else
	       {
	          ss += r;
			  if (ss >= vi_1j && ss >= vij_1 && ss<sol) sol = ss;
	       }
	     }
	     else { negd++; return; }								//should never happen, but still...
	  }
	  else sol = min(sol,1 + vi_1j);							//sol determined by one point: solve order-1 equation
	else        
	  if (fij_1 == FLAGS::ALIVE || fij_1 == FLAGS::EXTREMUM) 
		sol = min(sol,1 + vij_1);								//sol determined by one point: solve order-1 equation
}



int FastMarchingMethod::diffuse()
{
    static NewValue newp[20]; 

    //*** 1. FIND POINT IN NARROWBAND WITH LOWEST DISTANCE-VALUE
    int min_i,min_j;
    std::multimap<float,Coord>::iterator it=map.begin();
    if (it==map.end()) return 0;

    min_i  = (*it).second.i;
    min_j  = (*it).second.j;
    map.erase(it);					//erase point from 'map', since we'll make it alive in step 2

    //*** 2. MAKE MIN-POINT ALIVE
    (*flags)(min_j,min_i) = FLAGS::ALIVE;		
    if ((*f)(min_j,min_i)>=maxf) return 1;		//stop evolution if we reached the user-prescribed threshold.
   


     //*** 3. FIND ALL ITS STILL-TO-BE-UPDATED NEIGHBOURS
    Coord nbs[4]; int nn = 0;
    tag_nbs(min_i-1,min_j,min_i,min_j,nbs,nn);
    tag_nbs(min_i+1,min_j,min_i,min_j,nbs,nn);
    tag_nbs(min_i,min_j-1,min_i,min_j,nbs,nn);
    tag_nbs(min_i,min_j+1,min_i,min_j,nbs,nn);

    if (!nn) 						//only alive-neighbours of point (min_i,min_j) found,
    { 							//so it should be an extremum point...
		(*flags)(min_j,min_i) = FLAGS::EXTREMUM;
        nextr++;
		return 1;	
    }

    //*** 4. UPDATE VALUES OF NEIGHBOURS OF MIN-POINT
    NewValue* nnewp = newp;				//start updating neighbours. Their new values will be saved
    for(nn--;nn>=0;nn--)				//in nnewp[] and pasted back in 'f' at the update end.
    {
	int i = nbs[nn].i;
	int j = nbs[nn].j;

	float vi_1j = (*f)(j,i-1),     vijx1 = (*f)(j+1,i);
	float vix1j = (*f)(j,i+1),     vij_1 = (*f)(j-1,i);
	int   fi_1j = (*flags)(j,i-1), fijx1 = (*flags)(j+1,i);
	int   fix1j = (*flags)(j,i+1), fij_1 = (*flags)(j-1,i);

	float sol = FLAGS::MYINFINITY;
	solve(fi_1j,fij_1,vi_1j,vij_1,sol);
	solve(fix1j,fij_1,vix1j,vij_1,sol);
	solve(fi_1j,fijx1,vi_1j,vijx1,sol); 
	solve(fix1j,fijx1,vix1j,vijx1,sol); 

	if (sol < FLAGS::MYINFINITY/2) 
        { nnewp->i = i; nnewp->j = j; nnewp->value = sol; nnewp++; } 
    }

    //***5. Write updated values back in field.
    for(nnewp--;nnewp>=newp;nnewp--)				//for all updated neighbours:
    {
       map.erase(ptrs(nnewp->j,nnewp->i));		//remove the neighbour's entry from the sorted map...
       std::multimap<float,Coord>::value_type v(nnewp->value,Coord(nnewp->i,nnewp->j));
       ptrs(nnewp->j,nnewp->i) = map.insert(v);	//...and insert it back since its field-value changed		
       (*f)(nnewp->j,nnewp->i) = nnewp->value;		//update the field too!
   }
 
    return 1;
}







void FastMarchingMethod::tag_nbs(int min_i,int min_j,		//Gathers all neighbours (min_i,min_j) of (i,j)
                                 int i,int j,Coord* nbs,	//are to be updated because updating (i,j).
								 int& nn)
{
    if (min_i>=0 && min_i<flags->getWidth() && min_j>=0 && min_j<flags->getHeight() && !flags->alive(min_i,min_j))
    {
       nbs[nn].i = min_i; nbs[nn].j = min_j; nn++;
       if ((*flags)(min_j,min_i)!=FLAGS::NARROW_BAND)
          add_to_narrowband(min_i,min_j,i,j);			//Here we do our gathering stuff
    }
}



void FastMarchingMethod::add_to_narrowband(int i,int j,int,int)	//Adds point i,j to narrowband.
{
	(*flags)(j,i) = FLAGS::NARROW_BAND;
	ptrs(j,i)  = map.insert(std::make_pair((*f)(j,i),Coord(i,j)));
}



