#pragma once

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <assert.h>

typedef unsigned char byte;



struct RGBPixel								//Models a RGB pixel. Useful to create color images
{											//using Image<RGBPixel>.
			RGBPixel(float r_,float g_,float b_): r(r_),g(g_),b(b_) {}
			RGBPixel(float v): r(v),g(v),b(v) {}
			RGBPixel() {}

  RGBPixel& operator+=(const RGBPixel& p)
			{ r += p.r; g += p.g; b += p.b; return *this; }

  RGBPixel& operator-=(const RGBPixel& p)
			{ r -= p.r; g -= p.g; b -= p.b; return *this; }
			
  bool      operator<(const RGBPixel& p) const
			{ return r+g+b < p.r+p.g+p.b; }		

  float		r,g,b;		
};	




template<class T> class Image				//A simple class that represents a 2D image. The pixel type can be specified
{											//as the template parameter T.
 public:

  T**		pixels;

 public:
			Image();
			Image(int w, int h);
			Image(const Image<T> &);
  virtual  ~Image();
  int		getWidth() const { return width; }
  int		getHeight() const { return height; }
  int		getSize() { return width*height; }
  void		makeImage(int w, int h);
  T*		getPixels() { return pixels[0]; }
  void		setPixels(int w, int h, T *data);  
  T*		operator[](short i) { return pixels[i]; }
  const T*	operator[](short i) const { return pixels[i]; }
  T&		operator()(short i, short j) {  return pixels[0][i*width+j];  }
  const T&  operator()(short i, short j) const {  return pixels[0][i*width+j];  }
  Image<T>& operator=(const Image< T > &img); 
  int		operator==(const Image< T > &img);
  int		operator!=(const Image< T > &img);
  Image<T>& operator+=(const T val);
  Image<T>& operator-=(const T val);
  Image<T>& operator*=(const T val);
  Image<T>& operator+=(const Image< T > & img);
  Image<T>& operator*=(const Image< T > & img);
  Image<T>& operator-=(const Image< T > & img);
  void      clearImage(const T val);
  void      clearImage();
  void		range(T& min_val,T& max_val);
  
  static
  Image<T>* readPGM(const char* filename);

  static
  Image<T>* readPPM(const char* filename);

  void		writePGM(const char* filename) const;
  void		writePPM(const char* filename) const;

protected: 
  
  short		width, height; 
  void		destroyImage();
};



template< class T > Image< T >::Image()
{
  width = height = 0; 
  pixels = 0;
}

template< class T > Image< T >::Image(int w, int h)
{
    pixels = (T **)calloc(h, sizeof(T *));
    assert(pixels);
    pixels[0] = (T *)calloc(w*h, sizeof(T));
    assert(pixels[0]);

    for(int i=1;i<h;i++)
       pixels[i] = pixels[i-1] + w;

  width = w; height = h;
}

template< class T > Image< T >::Image(const Image<T>& img)
{
  width  = img.width;
  height = img.height;
  pixels = 0;
  if (img.width*img.height != 0) 
      setPixels(img.width, img.height, img.pixels[0]);
}

template< class T > void Image< T >::destroyImage() 
{
  if(pixels != 0) 
  {
    free(pixels[0]);
    free(pixels);
    pixels = 0;
    width=height=0;
  }
}

template< class T > Image< T >::~Image()
{ 
  destroyImage();
}

template< class T > void Image< T >::makeImage(int w, int h)
{
    destroyImage(); 

    pixels = (T **)calloc(h, sizeof(T*));
    assert(pixels);
    pixels[0] = (T*)calloc(w*h, sizeof(T));
    assert(pixels[0]);


    for(int i=1;i<h;i++)
      pixels[i] = pixels[i-1] + w;

    width = w; height = h;
}

template< class T > void Image< T >::setPixels(int w, int h, T* data)
{
    makeImage(w, h);

    assert(data);
    memcpy(pixels[0], data, getSize()*sizeof(T));
}


template< class T > Image< T >& Image< T >::operator=(const Image< T >& img)
{
  if(this != &img) 
  {
    width = img.width;
    height = img.height;
    if (img.width*img.height != 0)
      setPixels(img.width, img.height, img.pixels[0]);
    else
      pixels = 0;
  }
  return *this;
}

template< class T > int Image< T >::operator==(const Image< T >&img)
{
  if (width==img.width && height==img.height)
    return (!memcmp(pixels[0], img.pixels[0], getSize()*sizeof(T)));
  else
    return (width==img.width && height==img.height); 
}

template< class T > int Image< T >::operator!=(const Image< T >&img)
{
  if (width==img.width && height==img.height)
    return (memcmp(pixels[0], img.pixels[0], getSize()*sizeof(T)));
  else
    return (width!=img.width && height!=img.height); 
}

template< class T > Image< T > &Image< T >::operator+=(const T val)
{
  for (int i=0;i<getSize();i++) 
    pixels[0][i] += val;
  return (*this);
}

template< class T > Image< T > &Image< T >::operator-=(const T val)
{
  for (int i=0;i<getSize();i++) 
    pixels[0][i] -= val;
  return (*this);
}

template< class T > Image< T > &Image< T >::operator+=(const Image< T >&img)
{
  if(width!=img.width || height!=img.height)
    printf("Error: The two images have different sizes");
  else
  {
    for(int i=0;i<getSize();i++)
      pixels[0][i] += img.pixels[0][i];
    return *this;
  }
}

template< class T > Image< T > &Image< T >::operator-=(const Image< T >&img)
{
  if(width!=img.width || height!=img.height)
  {
    printf("Error: The two images have different sizes");
  }
  else
  {
    for(int i=0;i<getSize();i++)
      pixels[0][i] -= img.pixels[0][i];
    return *this;
  }
}

template< class T > void Image< T >::clearImage() 
{ 
  memset(pixels[0],0,sizeof(T)*getSize());
}

template< class T > void Image< T >::clearImage(const T val) 
{ 
  for (int i=0,sz=getSize(); i<sz; ++i)
    pixels[0][i] = val;
}



template< class T > Image<T>* Image< T >::readPGM(const char* filename) 
{
	FILE* fp = fopen(filename,"rb"); if (!fp) return 0;

   const int SIZE = 1024;
   char buf[SIZE]; int dimX,dimY,range;
   fscanf(fp,"%*s");									//skip "P5" header

   for(;;)
   {
     fscanf(fp,"%s",buf);								//get dimX or #comment
     if (buf[0]=='#') fgets(buf,SIZE,fp); 
        else { dimX = atoi(buf); break; }
   }
   for(;;)
   {
     fscanf(fp,"%s",buf);								//get dimY or #comment
     if (buf[0]=='#') fgets(buf,SIZE,fp); 
        else { dimY = atoi(buf); break; }
   }
   for(;;)
   {
     fscanf(fp,"%s",buf);								//get range or #comment
     if (buf[0]=='#') fgets(buf,SIZE,fp); 
        else { range = atoi(buf); break; }
   }
   

   Image<T>* f = new Image<T>(dimX,dimY);
   int bb = SIZE; fgets(buf,SIZE,fp);
  
   for(int j=0;j<dimY;++j)
   {
	  for(int i=0;i<dimX;++i)							//read the binary data into the image
	  {													//do not modify the range of the read image
		if (bb==SIZE) { fread(buf,SIZE,1,fp); bb=0; }
		(*f)(j,i) = (unsigned char)buf[bb++];
	  }	
   }

   fclose(fp);  				
   return f;
}


template< class T > Image<T>* Image< T >::readPPM(const char* filename) 
{
	FILE* fp = fopen(filename,"rb"); if (!fp) return 0;

   const int SIZE = 3*1024;
   char buf[SIZE]; int dimX,dimY,range;
   fscanf(fp,"%*s");									//skip "P6" header

   for(;;)
   {
     fscanf(fp,"%s",buf);								//get dimX or #comment
     if (buf[0]=='#') fgets(buf,SIZE,fp); 
        else { dimX = atoi(buf); break; }
   }
   for(;;)
   {
     fscanf(fp,"%s",buf);								//get dimY or #comment
     if (buf[0]=='#') fgets(buf,SIZE,fp); 
        else { dimY = atoi(buf); break; }
   }
   for(;;)
   {
     fscanf(fp,"%s",buf);								//get range or #comment
     if (buf[0]=='#') fgets(buf,SIZE,fp); 
        else { range = atoi(buf); break; }
   }
   

   Image<T>* f = new Image<T>(dimX,dimY);
   int bb = SIZE; fgets(buf,SIZE,fp);
  
   for(int j=0;j<dimY;++j)
   {
	  for(int i=0;i<dimX;++i)							//read the binary data into the image
	  {													//do not modify the range of the image
		if (bb==SIZE) { fread(buf,SIZE,1,fp); bb=0; }
		
		float r = (unsigned char)buf[bb++];
		float g = (unsigned char)buf[bb++];
		float b = (unsigned char)buf[bb++];
														//since we do not have a 3-channel image,
		(*f)(j,i) = (r+g+b)/3;							//store just the luminance of the read color-pixel	
	  }	
   }

   fclose(fp);  				
   return f;
}


template< class T > void Image< T >::writePGM(const char* fname) const
{
   FILE* fp = fopen(fname,"wb");
   if (!fp) return;

   const int SIZE = 1024;
   unsigned char buf[SIZE];
   int bb=0;

   fprintf(fp,"P5 %d %d 255\n",getWidth(),getHeight());
   
   for(int j=0;j<getHeight();++j)
   {
	  for(int i=0;i<getWidth();++i)						//
	  {													//
	     float v = (*this)(j,i);
         buf[bb++] = (unsigned char)(int)(v);
		 if (bb==SIZE)
         {  fwrite(buf,sizeof(unsigned char),SIZE,fp); bb = 0; }
	  }	 
   }
   if (bb) fwrite(buf,sizeof(unsigned char),bb,fp);

   fclose(fp);
}	

template< class T > void Image< T >::writePPM(const char* fname) const
{
   FILE* fp = fopen(fname,"wb");
   if (!fp) return;

   const int SIZE = 3*1024;
   unsigned char buf[SIZE];
   int bb=0;

   fprintf(fp,"P6 %d %d 255\n",getWidth(),getHeight());
   
   for(int j=0;j<getHeight();++j)
   {
	  for(int i=0;i<getWidth();++i)						//
	  {													//
	     float v = (*this)(j,i);
         buf[bb++] = (unsigned char)(int)(v);
         buf[bb++] = (unsigned char)(int)(v);
         buf[bb++] = (unsigned char)(int)(v);

		 if (bb==SIZE)
         {  fwrite(buf,sizeof(unsigned char),SIZE,fp); bb = 0; }
	  }	 
   }
   if (bb) fwrite(buf,sizeof(unsigned char),bb,fp);

   fclose(fp);
}	




template< class T> void Image< T >::range(T& min_val,T& max_val)
{
	min_val = (*this)(0,0);
	max_val = (*this)(0,0);

	for(int j=0;j<getHeight();++j)
	  for(int i=0;i<getWidth();++i)
	  {
	    const T value = (*this)(j,i);
		min_val = std::min(min_val,value);
		max_val = std::max(max_val,value);
	  }
}




